@@ -1,18 +1,81 @@
-import numpy as np
 import random
 import subprocess
-
-
+import numpy as np
 from .fasta import read_fasta
 from .dnatools import rev_comp
-from .jellyfish import jellyfish_count
+from .lowlevel import call, _temp, gzip, printf
+from .file_conversion import fq_to_fa
+
+
+def jellyfish_count(file, k, output, min_count=1, hash_size='100M', threads=4, canonical=True):
+    """
+    Wrapper function for the jellyfish k-mer count commands
+
+    From a high-level it takes an input fastq or fastq <file>, and eventually
+        exports an <output> fasta file of k-mer counts
+
+    All the intermediate files are deleted
+
+    Args:
+        file: str, path-like
+            The input fasta or fastq file, also accepts .gz compressed format
+
+        k: int
+            The size of the k-mer
+
+        output: str, path-like
+            The output fasta file, which cannot be the same as the input <file>
+
+        min_count: uint
+            K-mers with counts < min_count will be excluded
+
+        hash_size: str,
+            e.g. '100M' for 100 million; '1G' for 1 billion
+
+        threads: int (uint32)
+
+        canonical: bool
+            If True, 'AAAC' will be the same as 'GTTT'
+    """
+    # The <output> fasta should NOT be identical to the input fasta <file>
+    if output == file: return
+
+    input_name = file  # The original input name
+
+    # Unzip
+    if file.endswith('.gz'):
+        gzip(file, keep=True)
+        file = file[:-3]
+
+    # Convert fastq to fasta
+    if file.endswith('.fastq') or file.endswith('.fq'):
+        fq_to_fa(file, keep=True)
+        # If the input <file> is not equal to the original input name, then don't keep it
+        if not file == input_name: call(f"rm {file}")
+
+    if file.endswith('.fastq'): file = file[:-6] + '.fa'
+    elif file.endswith('.fq'): file = file[:-3] + '.fa'
+
+    # Count k-mers in the input fastq or fasta file
+    canonical = ['', '-C '][canonical]
+    temp_jf = _temp('temp', '.jf')  # for example, temp000.jf
+    call(f"jellyfish count -m {k} -s {hash_size} -t {threads} -o {temp_jf} {canonical}{file}")
+
+    # Remove the counted file if it's not the original input file
+    if not file == input_name: call(f"rm {file}")
+
+    # Dump the .jf file into a .fa file
+    call(f"jellyfish dump -L {min_count} --output {output} {temp_jf}")
+
+    # The .jf file is not needed, so remove it.
+    call(f"rm {temp_jf}")
 
 
 def read_kmers(jf_fa, min_count=1):
     """
     Args:
-        jf_fa: str
-            Name of fasta exported by jellyfish
+        jf_fa: str, path-like
+            The fasta exported by jellyfish
 
         min_count: int
             Kmers >= min_count will be included
@@ -43,11 +106,11 @@ def read_kmers(jf_fa, min_count=1):
 def kmer_2D_points(jf_fa_1, jf_fa_2, min_count=1):
     """
     Args:
-        jf_fa_1: str
-            Name of fasta exported by jellyfish
+        jf_fa_1: str, path-like
+            The fasta exported by jellyfish
 
-        jf_fa_2: str
-            Name of fasta exported by jellyfish
+        jf_fa_2: str, path-like
+            The fasta exported by jellyfish
 
         min_count: int
             Kmers >= min_count will be included from the jellyfish fasta
@@ -119,7 +182,7 @@ def save_kmer_2D_points(kmers, x, y, file, x_name='x', y_name='y', description='
     """
     This method writes a tab-delimited file to store 2D points of kmers.
 
-    The first line is the comment/description text.
+    The first line is the description text.
 
     The second line is the header line:
         'k-mer' \t <x_name> \t <y_name>
@@ -128,7 +191,7 @@ def save_kmer_2D_points(kmers, x, y, file, x_name='x', y_name='y', description='
         <kmer_seq> \t <x_coordinate> \t <y_coordinate>
 
     Here's an example:
-               \t <x_name> \t <y_name>
+        k-mer  \t <x_name> \t <y_name>
         AAAAAA \t 10       \t 10
         AAAAAC \t 1        \t 1
 
@@ -159,10 +222,10 @@ def save_kmer_2D_points(kmers, x, y, file, x_name='x', y_name='y', description='
     Returns: None
     """
     if len(kmers) != len(x):
-        print("The number of 'kmers' is not equal to that of 'x'")
+        printf("The number of 'kmers' is not equal to that of 'x'")
         return
     if len(kmers) != len(y):
-        print("The number of 'kmers' is not equal to that of 'y'")
+        printf("The number of 'kmers' is not equal to that of 'y'")
         return
 
     if not file.endswith('.tsv'):
@@ -175,12 +238,12 @@ def save_kmer_2D_points(kmers, x, y, file, x_name='x', y_name='y', description='
 
     with open(file, 'w') as fh:
         fh.write(description + '\n')
-        fh.write('\t{}\t{}\n'.format(x_name, y_name))
+        fh.write('k-mer\t{}\t{}\n'.format(x_name, y_name))
         for i in range(len(kmers)):
             fh.write('{}\t{}\t{}\n'.format(kmers[i], x[i], y[i]))
 
 
-def read_kmer_2D_points(file, dtype=np.uint32):
+def read_kmer_2D_points(file, dtype=np.uint32, fraction=1.):
     """
     This function reads the tab-delimited file written by the method <save_kmer_2D_points>.
 
@@ -193,7 +256,7 @@ def read_kmer_2D_points(file, dtype=np.uint32):
         <kmer_seq> \t <x_coordinate> \t <y_coordinate>
 
     Here's an example:
-               \t <x_name> \t <y_name>
+        k-mer  \t <x_name> \t <y_name>
         AAAAAA \t 10       \t 10
         AAAAAC \t 1        \t 1
 
@@ -204,6 +267,10 @@ def read_kmer_2D_points(file, dtype=np.uint32):
         dtype: numpy data types
             Number data types for x and y coordinates; np.uint32 is the default
 
+        fraction: float
+            The fraction of k-mers to be randomly retrieved from the file
+            Default = 1., i.e. all k-mers
+
     Returns:
         three arrays:
             kmers: list of str
@@ -233,11 +300,12 @@ def read_kmer_2D_points(file, dtype=np.uint32):
         description = fh.readline().rstrip()
         _, x_name, y_name = fh.readline().rstrip().split('\t')
         for line in fh:
-            data.append(
-                tuple(
-                    line.rstrip().split('\t') # 'kmer\tx\ty' => (kmer, x, y)
+            if random.uniform(0, 1) <= fraction:
+                data.append(
+                    tuple(
+                        line.rstrip().split('\t') # 'kmer\tx\ty' => (kmer, x, y)
+                    )
                 )
-            )
 
     kmers, x, y = zip(*data) # Unzip into three lists
 
@@ -487,8 +555,11 @@ def get_kmer_2D_points(seq, kmer_dict):
         elif kmer_dict.get(rc_kmr, None):
             seq_kmer_dict.setdefault(rc_kmr, kmer_dict[rc_kmr])
 
-    kmers, xy = zip(*seq_kmer_dict.items())
-    x, y = zip(*xy)
+    if seq_kmer_dict:
+        kmers, xy = zip(*seq_kmer_dict.items())
+        x, y = zip(*xy)
+    else:
+        kmers, x, y = [], [], []
 
     return kmers, np.array(x), np.array(y)
 
@@ -515,9 +586,10 @@ def kmers_base_content(kmers, base):
 def fastq_to_saved_kmer_2D_points(fq1, fq2, k, output, min_count=1, hash_size='100M', threads=4, description=''):
     """
     The full implementation of the pipeline:
-        Two fastq files
-            -> k-mer counts of each fastq file
-                -> tab-separated file of k-mer 2D points
+        (1) Input two fastq files
+        (2) Count k-mers in each fastq file
+        (3) Merge (outer join) the k-mer counts of the two files
+        (4) Write a tab-separated file (.tsv) of k-mer 2D points
 
     Args:
         fq1: str, path-like
@@ -556,4 +628,3 @@ def fastq_to_saved_kmer_2D_points(fq1, fq2, k, output, min_count=1, hash_size='1
 
     # Remove the two fasta files exported by jellyfish
     subprocess.check_call('rm 1.fa 2.fa', shell=True)
-
